import datetime

from django import forms
from django.utils import six

from .helpers import nice_repr, parse

class TimedeltaWidget(forms.TextInput):
    def __init__(self, *args, **kwargs):
        return super(TimedeltaWidget, self).__init__(*args, **kwargs)

    def render(self, name, value, attrs=None):
        if value is None:
            value = ""
        elif isinstance(value, six.string_types):
            pass
        else:
            if isinstance(value, int):
                value = datetime.timedelta(seconds=value)
            value = nice_repr(value)
        return super(TimedeltaWidget, self).render(name, value, attrs)

    def _has_changed(self, initial, data):
        """
        We need to make sure the objects are of the canonical form, as a
        string comparison may needlessly fail.
        """
        if initial in ["", None] and data in ["", None]:
            return False

        if initial in ["", None] or data in ["", None]:
            return True

        if initial:
            if not isinstance(initial, datetime.timedelta):
                initial = parse(initial)

        if not isinstance(data, datetime.timedelta):
            try:
                data = parse(data)
            except TypeError:
                # initial didn't throw a TypeError, so this must be different
                # from initial
                return True

        return initial != data



from django.db import models
from django.core.exceptions import ValidationError
from django.utils import six

from collections import defaultdict
import datetime

from .helpers import parse
from .forms import TimedeltaFormField

# TODO: Figure out why django admin thinks fields of this type have changed every time an object is saved.

# Define the different column types that different databases can use.
COLUMN_TYPES = defaultdict(lambda:"char(20)")
COLUMN_TYPES["django.db.backends.postgresql_psycopg2"] = "interval"
COLUMN_TYPES["django.contrib.gis.db.backends.postgis"] = "interval"

class TimedeltaField(six.with_metaclass(models.SubfieldBase, models.Field)):
    """
    Store a datetime.timedelta as an INTERVAL in postgres, or a
    CHAR(20) in other database backends.
    """
    _south_introspects = True

    description = "A datetime.timedelta object"

    def __init__(self, *args, **kwargs):
        self._min_value = kwargs.pop('min_value', None)

        if isinstance(self._min_value, (int, float)):
            self._min_value = datetime.timedelta(seconds=self._min_value)

        self._max_value = kwargs.pop('max_value', None)

        if isinstance(self._max_value, (int, float)):
            self._max_value = datetime.timedelta(seconds=self._max_value)

        super(TimedeltaField, self).__init__(*args, **kwargs)

    def to_python(self, value):
        if (value is None) or isinstance(value, datetime.timedelta):
            return value
        if isinstance(value, (int, float)):
            return datetime.timedelta(seconds=value)
        if value == "":
            if self.null:
                return None
            else:
                return datetime.timedelta(0)
        return parse(value)

    def get_prep_value(self, value):
        if self.null and value == "":
            return None
        if (value is None) or isinstance(value, six.string_types):
            return value
        return str(value).replace(',', '')

    def get_db_prep_value(self, value, connection=None, prepared=None):
        return self.get_prep_value(value)

    def formfield(self, *args, **kwargs):
        defaults = {'form_class':TimedeltaFormField}
        defaults.update(kwargs)
        return super(TimedeltaField, self).formfield(*args, **defaults)

    def validate(self, value, model_instance):
        super(TimedeltaField, self).validate(value, model_instance)
        if self._min_value is not None:
            if self._min_value > value:
                raise ValidationError('Less than minimum allowed value')
        if self._max_value is not None:
            if self._max_value < value:
                raise ValidationError('More than maximum allowed value')

    def value_to_string(self, obj):
        value = self._get_val_from_obj(obj)
        return unicode(value)

    def get_default(self):
        """
        Needed to rewrite this, as the parent class turns this value into a
        unicode string. That sux pretty deep.
        """
        if self.has_default():
            if callable(self.default):
                return self.default()
            return self.get_prep_value(self.default)
        if not self.empty_strings_allowed or (self.null):
            return None
        return ""

    def db_type(self, connection):
        return COLUMN_TYPES[connection.settings_dict['ENGINE']]

    def deconstruct(self):
        """
        Break down this field into arguments that can be used to reproduce it
        with Django migrations.

        The thing to to note here is that currently the migration file writer
        can't serialize timedelta objects so we convert them to a float
        representation (in seconds) that we can later interpret as a timedelta.
        """

        name, path, args, kwargs = super(TimedeltaField, self).deconstruct()

        if isinstance(self._min_value, datetime.timedelta):
            kwargs['min_value'] = self._min_value.total_seconds()

        if isinstance(self._max_value, datetime.timedelta):
            kwargs['max_value'] = self._max_value.total_seconds()

        if isinstance(kwargs.get('default'), datetime.timedelta):
            kwargs['default'] = kwargs['default'].total_seconds()

        return name, path, args, kwargs



from unittest import TestCase, skipIf
import datetime
import doctest

import django
from django.core.exceptions import ValidationError
from django.db import models

from .fields import TimedeltaField
import timedelta.helpers
import timedelta.forms
import timedelta.widgets


class MinMaxTestModel(models.Model):
    min = TimedeltaField(min_value=datetime.timedelta(1))
    max = TimedeltaField(max_value=datetime.timedelta(1))
    minmax = TimedeltaField(min_value=datetime.timedelta(1), max_value=datetime.timedelta(7))


class IntTestModel(models.Model):
    field = TimedeltaField(min_value=1, max_value=86400)


class FloatTestModel(models.Model):
    field = TimedeltaField(min_value=1.0, max_value=86400.0)


class TimedeltaModelFieldTest(TestCase):
    def test_validate(self):
        test = MinMaxTestModel(
            min=datetime.timedelta(1),
            max=datetime.timedelta(1),
            minmax=datetime.timedelta(1)
        )
        test.full_clean()  # This should have met validation requirements.

        test.min = datetime.timedelta(hours=23)
        self.assertRaises(ValidationError, test.full_clean)

        test.min = datetime.timedelta(hours=25)
        test.full_clean()

        test.max = datetime.timedelta(11)
        self.assertRaises(ValidationError, test.full_clean)

        test.max = datetime.timedelta(hours=20)
        test.full_clean()

        test.minmax = datetime.timedelta(0)
        self.assertRaises(ValidationError, test.full_clean)
        test.minmax = datetime.timedelta(22)
        self.assertRaises(ValidationError, test.full_clean)
        test.minmax = datetime.timedelta(6, hours=23, minutes=59, seconds=59)
        test.full_clean()

    def test_from_int(self):
        """
        Check that integers can be used to define the min_value and max_value
        arguments, and that when assigned an integer, TimedeltaField converts
        to timedelta.
        """

        test = IntTestModel()

        # valid
        test.field = 3600
        self.assertEquals(test.field, datetime.timedelta(seconds=3600))
        test.full_clean()

        # invalid
        test.field = 0
        self.assertRaises(ValidationError, test.full_clean)

        # also invalid
        test.field = 86401
        self.assertRaises(ValidationError, test.full_clean)

    def test_from_float(self):
        """
        Check that floats can be used to define the min_value and max_value
        arguments, and that when assigned a float, TimedeltaField converts
        to timedelta.
        """

        test = FloatTestModel()

        # valid
        test.field = 3600.0
        self.assertEquals(test.field, datetime.timedelta(seconds=3600))
        test.full_clean()

        # invalid
        test.field = 0.0
        self.assertRaises(ValidationError, test.full_clean)

        # also invalid
        test.field = 86401.0
        self.assertRaises(ValidationError, test.full_clean)

    @skipIf(django.VERSION < (1, 7), 'deconstruct not present before 1.7')
    def test_deconstruct(self):
        """
        Check that the deconstruct() method of TimedeltaField is returning the
        min_value, max_value and default kwargs as floats.
        """

        field = TimedeltaField(
            min_value=datetime.timedelta(minutes=5),
            max_value=datetime.timedelta(minutes=15),
            default=datetime.timedelta(minutes=30),
        )

        kwargs = field.deconstruct()[3]
        self.assertEqual(kwargs['default'], 1800.0)
        self.assertEqual(kwargs['max_value'], 900.0)
        self.assertEqual(kwargs['min_value'], 300.0)

    def test_load_from_db(self):
        obj = MinMaxTestModel.objects.create(min='2 days', max='2 minutes', minmax='3 days')
        self.assertEquals(datetime.timedelta(2), obj.min)
        self.assertEquals(datetime.timedelta(0, 120), obj.max)
        self.assertEquals(datetime.timedelta(3), obj.minmax)

        obj = MinMaxTestModel.objects.get()
        self.assertEquals(datetime.timedelta(2), obj.min)
        self.assertEquals(datetime.timedelta(0, 120), obj.max)
        self.assertEquals(datetime.timedelta(3), obj.minmax)


def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(timedelta.helpers))
    tests.addTests(doctest.DocTestSuite(timedelta.forms))
    return tests



from django import forms
from django.utils.translation import ugettext_lazy as _
from django.utils import six

import datetime
from collections import defaultdict

from .widgets import TimedeltaWidget
from .helpers import parse

class TimedeltaFormField(forms.Field):
    
    default_error_messages = {
        'invalid':_('Enter a valid time span: e.g. "3 days, 4 hours, 2 minutes"')
    }
    
    def __init__(self, *args, **kwargs):
        defaults = {'widget':TimedeltaWidget}
        defaults.update(kwargs)
        super(TimedeltaFormField, self).__init__(*args, **defaults)
        
    def clean(self, value):
        """
        This doesn't really need to be here: it should be tested in
        parse()...
        
        >>> t = TimedeltaFormField()
        >>> t.clean('1 day')
        datetime.timedelta(1)
        >>> t.clean('1 day, 0:00:00')
        datetime.timedelta(1)
        >>> t.clean('1 day, 8:42:42.342')
        datetime.timedelta(1, 31362, 342000)
        >>> t.clean('3 days, 8:42:42.342161')
        datetime.timedelta(3, 31362, 342161)
        >>> try:
        ...  t.clean('3 days, 8:42:42.3.42161')
        ... except forms.ValidationError as arg:
        ...  six.print_(arg.messages[0])
        Enter a valid time span: e.g. "3 days, 4 hours, 2 minutes"
        >>> t.clean('5 day, 8:42:42')
        datetime.timedelta(5, 31362)
        >>> t.clean('1 days')
        datetime.timedelta(1)
        >>> t.clean('1 second')
        datetime.timedelta(0, 1)
        >>> t.clean('1 sec')
        datetime.timedelta(0, 1)
        >>> t.clean('10 seconds')
        datetime.timedelta(0, 10)
        >>> t.clean('30 seconds')
        datetime.timedelta(0, 30)
        >>> t.clean('1 minute, 30 seconds')
        datetime.timedelta(0, 90)
        >>> t.clean('2.5 minutes')
        datetime.timedelta(0, 150)
        >>> t.clean('2 minutes, 30 seconds')
        datetime.timedelta(0, 150)
        >>> t.clean('.5 hours')
        datetime.timedelta(0, 1800)
        >>> t.clean('30 minutes')
        datetime.timedelta(0, 1800)
        >>> t.clean('1 hour')
        datetime.timedelta(0, 3600)
        >>> t.clean('5.5 hours')
        datetime.timedelta(0, 19800)
        >>> t.clean('1 day, 1 hour, 30 mins')
        datetime.timedelta(1, 5400)
        >>> t.clean('8 min')
        datetime.timedelta(0, 480)
        >>> t.clean('3 days, 12 hours')
        datetime.timedelta(3, 43200)
        >>> t.clean('3.5 day')
        datetime.timedelta(3, 43200)
        >>> t.clean('1 week')
        datetime.timedelta(7)
        >>> t.clean('2 weeks, 2 days')
        datetime.timedelta(16)
        >>> try:
        ...  t.clean(six.u('2 we\xe8k, 2 days'))
        ... except forms.ValidationError as arg:
        ...  six.print_(arg.messages[0])
        Enter a valid time span: e.g. "3 days, 4 hours, 2 minutes"
        """
        
        super(TimedeltaFormField, self).clean(value)
        if value == '' and not self.required:
            return ''
        try:
            return parse(value)
        except TypeError:
            raise forms.ValidationError(self.error_messages['invalid'])

class TimedeltaChoicesField(TimedeltaFormField):
    def __init__(self, *args, **kwargs):
        choices = kwargs.pop('choices')
        defaults = {'widget':forms.Select(choices=choices)}
        defaults.update(kwargs)
        super(TimedeltaChoicesField, self).__init__(*args, **defaults)



import os.path
from setuptools import setup

setup(
    name="django-timedeltafield",
    version=open(os.path.join(os.path.dirname(__file__), 'timedelta', 'VERSION')).read().strip(),
    description="TimedeltaField for django models",
    long_description=open("README").read(),
    url="http://bitbucket.org/schinckel/django-timedelta-field/",
    author="Matthew Schinckel",
    author_email="matt@schinckel.net",
    packages=[
        "timedelta",
        "timedelta.templatetags",
    ],
    package_data={'timedelta': ['VERSION']},
    classifiers=[
        'Framework :: Django :: 1.4',
        'Framework :: Django :: 1.5',
        'Framework :: Django :: 1.6',
        'Framework :: Django :: 1.7',
        'Framework :: Django :: 1.8',
        'Framework :: Django',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    test_suite='tests.main',
)
